extends KinematicBody2D
class_name Player

enum STATES{SPAWNING, IDLE, WALK, DEAD, HITTED}

var id = "mage"
var level:int = 1
var experience = 0
var next_level_exp = 0
var coins = 99999999
var upgrade_points = 99999999
var spells = {}
var items = {}

var accerelation:int = 40
var friction:float = 0.15

var direction:Vector2 = Vector2.ZERO
var last_direction = Vector2.ZERO
var velocity:Vector2 = Vector2.ZERO

var knockback_force = 0
var knockback_direction = Vector2.ZERO
var dashing = false

var state = STATES.SPAWNING

onready  var stats:Stats = $stats
onready  var sprite = $animated_sprite
onready  var shadow = $shadow
onready  var pick_area_shape = $pick_area / collision_shape_2d.shape
onready  var casters_container = $casters_container
onready  var input_controller = $input_controller
onready  var hurt_box_collision = $hurt_box / collision_shape_2d
onready  var collision = $collision_shape_2d


func _ready():
	randomize()
	Global.connect("enemy_died", self, "_on_enemy_died")
	
	Global.char_set_data(self, Database.get("char_" + id))
	set_level(1)

	
	_sync_stats_update()


func _process(_delta):
	if state == STATES.DEAD or state == STATES.SPAWNING:
		return 
	direction = input_controller.get_value()

	if direction:
		last_direction = direction
		sprite.play("walk")
	elif not dashing:
		sprite.play("idle")

	if direction.x != 0:
		sprite.flip_h = direction.x < 0
		shadow.position.x = 1 if sprite.flip_h else 0
	

func _physics_process(_delta:float)->void :
	if state == STATES.DEAD or state == STATES.SPAWNING:
		return 

	_move()


func set_level(value:int):
	level = value
	next_level_exp = _get_level_exp()
	Global.log_player_level(level)


func die():
	if state == STATES.DEAD:
		return 
	state = STATES.DEAD
	Global.emit_signal("player_deaded")
	queue_free()

	
func add_coins(_coins:int):
	coins += _coins
	Global.emit_signal("player_coins_changed", _coins)


func add_item(item_id:String):
	if items.has(item_id):
		return 
	items[item_id] = Entities.create_item_data(item_id)
	items[item_id].level = 1
	stats.add_modifier(items[item_id])
	Global.log_item_level(item_id, 1)
	Global.hud_item_slots.add(items[item_id])
	apply_modifiers()


func add_consumable(data):
	if "modifier_duration" in data:
		if not items.has(data.id):
			items[data.id] = data.duplicate()
			stats.add_modifier(items[data.id], true)
		else :
			stats.add_modifier(data, true)
		apply_modifiers()
	else :
		stats.apply_consumables([data])

	
	if data.id == "consumable_collector_orb":
		for index in Global.drop_container.get_child_count():
			Global.drop_container.get_child(index).set_picker(self, 3)
	elif data.id == "consumable_stats_refresher_orb":
		reset_stats_attrs()
		apply_modifiers()
		Global.emit_signal("player_upgrade_points_changed")
	elif data.id == "consumable_items_refresher_orb":
		reset_items()
		Global.emit_signal("player_coins_changed", 0)
		apply_modifiers()
	elif data.id == "consumable_spells_refresher_orb":
		reset_spells()
		Global.emit_signal("player_upgrade_points_changed")
		apply_modifiers()

	
	elif "current_health" in data or "max_health" in data:
		Global.emit_signal("player_health_changed")


func add_spell(spell_id:String):
	if spells.has(spell_id):
		return 
	spells[spell_id] = casters_container.add_caster(spell_id)


func remove_spell(spell_id:String):
	spells.erase(spell_id)
	casters_container.remove_caster(spell_id)


func remove_item(item_id:String):
	stats.remove_modifier("item_" + item_id)
	Global.hud_item_slots.release_slot(items[item_id])
	
	items.erase(item_id)


func upgrade_spell(spell_id:String, next_level = null):
	var caster = spells[spell_id]
	caster.upgrade(next_level)
	Global.log_spell_level(spell_id, caster.get_data().level)


func upgrade_item(item_id:String, next_level = null):
	var data = items[item_id]
	if not next_level:
		next_level = data.level + 1
	data.upgrade(next_level)
		
	var icon_node = Global.hud_item_slots.get_node_from_data(data)
	icon_node.get_node("label").text = str(data.level)
	
	Global.log_item_level(item_id, data.level)
	apply_modifiers()



func get_closest_target(position_find = global_position, ignore = [], distance = stats.attack_range):
	return Targets.get_closest_target(position_find, ignore, distance)
									

func get_random_targets(size = 1):
	return Targets.get_random_targets(global_position, [], size, stats.attack_range)
	

func get_random_target():
	return Targets.get_random_target(global_position, [], stats.attack_range)


func get_random_attack_range_position():
	var target_area = global_position.normalized() * stats.attack_range
	var x = rand_range( - target_area.x, target_area.x)
	var y = rand_range( - target_area.y, target_area.y)
	return global_position + Vector2(x, y)


func reset_stats_attrs():
	for key in Global.STATS_ATTRS:
		upgrade_points += stats.raw_value[key]
		stats.set_raw_value(key, 0)


func reset_items():
	for item_id in items.keys():
		if item_id.begins_with("consumable"):
			continue
		var player_data = items[item_id]
		var item_raw_data = Database.get("item_" + item_id)
		var upgrade_index = player_data.level - 2
		coins += item_raw_data.cost
		while upgrade_index >= 0:
			var upgrade_data = item_raw_data.upgrades[upgrade_index]
			coins += upgrade_data.cost
			upgrade_index -= 1
		remove_item(item_id)


func reset_spells():
	for spell_id in spells.keys():
		var spell_data = spells[spell_id].get_data()
		upgrade_points += spell_data.level * 10
		remove_spell(spell_id)


func _move():
	if knockback_force:
		velocity = knockback_direction * knockback_force
		knockback_force = lerp(knockback_force, 0, 0.1)
		if knockback_force < 10:
			knockback_force = 0
	
	if ( not dashing and not knockback_force) or direction:
		velocity += direction * accerelation
	elif dashing and not direction:
		velocity += last_direction * accerelation
		
	if not knockback_force:
		velocity = velocity.clamped(stats.move_speed)
	
	velocity = move_and_slide(velocity)
	velocity = lerp(velocity, Vector2.ZERO, friction)


func apply_modifiers():
	
	stats.apply_modifiers()


func _sync_stats_update():
	
	sprite.frames.set_animation_speed("walk", float(stats.move_speed) / float(stats.raw_value.move_speed) * 10)
	pick_area_shape.radius = stats.pick_area * 0.5
	dashing = stats.modifiers.has("dash")

	
	if stats.modifiers.has("absolute_chaos"):
		sprite.material.set_shader_param("border_color", Color("#78329b"))
		sprite.material.set_shader_param("add_border", true)
	else :
		sprite.material.set_shader_param("add_border", false)


func _get_level_exp():
	if level < 5:
		return ceil(level * 30)
	elif level < 10:
		return ceil(level * 50)
	elif level < 15:
		return ceil(level * 100)
	elif level < 20:
		return ceil(level * 200)
	return ceil(level * 400)


func _on_enemy_died(enemy:Node2D):
	experience += enemy.experience * stats.experience_factor
	var levelup = false
	while experience >= next_level_exp:
		levelup = true
		upgrade_points += 10
		experience -= next_level_exp
		set_level(level + 1)

	if levelup:
		Global.emit_signal("player_level_changed")
		Global.emit_signal("player_upgrade_points_changed")
		call_deferred("add_child_below_node", stats, Global.create_levelup_effect())
		call_deferred("add_child", Global.create_levelup_text())
	
	Global.emit_signal("player_exp_changed")


func _on_stats_deaded():
	die()


func _on_stats_hitted(result:Dictionary):
	knockback_direction = (result.position - global_position).normalized() * - 1
	knockback_force = result.damage_knockback
	Global.log_player_damage_taken(result.damage)


func _on_stats_health_changed():
	Global.emit_signal("player_health_changed")


func _on_stats_modifiers_applied():
	_sync_stats_update()
	

func _on_stats_modifier_removed(data):
	
	items.erase(data.id)
